/*
   This file is part of TALER
   Copyright (C) 2022 Taler Systems SA

   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 3, or (at your option) any later version.

   TALER is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
   A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along with
   TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
 */
/**
 * @file iterate_historic_reserve_revenue.c
 * @brief Low-level (statement-level) Postgres database access for the exchange
 * @author Christian Grothoff
 */
#include "taler/taler_pq_lib.h"
#include "auditor-database/iterate_historic_reserve_revenue.h"
#include "pg_helper.h"


/**
 * Hard upper bound on the number of records returned by a single
 * call, regardless of the limit requested by the client.
 */
#define MAX_RECORDS 50000


/**
 * Closure for #historic_reserve_revenue_cb().
 */
struct HistoricReserveRevenueContext
{
  /**
   * Function to call for each result.
   */
  TALER_AUDITORDB_HistoricReserveRevenueDataCallback cb;

  /**
   * Closure for @e cb.
   */
  void *cb_cls;

  /**
   * Plugin context.
   */
  struct TALER_AUDITORDB_PostgresContext *pg;

  /**
   * Number of results processed.
   */
  enum GNUNET_DB_QueryStatus qs;
};


/**
 * Helper function for #TALER_AUDITORDB_iterate_historic_reserve_revenue().
 * To be called with the results of a SELECT statement
 * that has returned @a num_results results.
 *
 * @param cls closure of type `struct HistoricRevenueContext *`
 * @param result the postgres result
 * @param num_results the number of results in @a result
 */
static void
historic_reserve_revenue_cb (void *cls,
                             PGresult *result,
                             unsigned int num_results)
{
  struct HistoricReserveRevenueContext *hrc = cls;
  struct TALER_AUDITORDB_PostgresContext *pg = hrc->pg;

  for (unsigned int i = 0; i < num_results; i++)
  {
    uint64_t rowid;
    struct GNUNET_TIME_Timestamp start_date;
    struct GNUNET_TIME_Timestamp end_date;
    struct TALER_Amount reserve_profits;
    struct GNUNET_PQ_ResultSpec rs[] = {
      GNUNET_PQ_result_spec_uint64 ("row_id",
                                    &rowid),
      GNUNET_PQ_result_spec_timestamp ("start_date",
                                       &start_date),
      GNUNET_PQ_result_spec_timestamp ("end_date",
                                       &end_date),
      TALER_PQ_RESULT_SPEC_AMOUNT ("reserve_profits",
                                   &reserve_profits),
      GNUNET_PQ_result_spec_end
    };

    if (GNUNET_OK !=
        GNUNET_PQ_extract_result (result,
                                  rs,
                                  i))
    {
      GNUNET_break (0);
      hrc->qs = GNUNET_DB_STATUS_HARD_ERROR;
      return;
    }
    hrc->qs = i + 1;
    if (GNUNET_OK !=
        hrc->cb (hrc->cb_cls,
                 rowid,
                 start_date,
                 end_date,
                 &reserve_profits))
      break;
  }
}


enum GNUNET_DB_QueryStatus
TALER_AUDITORDB_iterate_historic_reserve_revenue (
  struct TALER_AUDITORDB_PostgresContext *pg,
  int64_t limit,
  uint64_t offset,
  TALER_AUDITORDB_HistoricReserveRevenueDataCallback cb,
  void *cb_cls)
{
  uint64_t ulimit = GNUNET_MIN ((uint64_t) MAX_RECORDS,
                                TALER_AUDITORDB_abs_limit (limit));
  struct GNUNET_PQ_QueryParam params[] = {
    GNUNET_PQ_query_param_uint64 (&offset),
    GNUNET_PQ_query_param_uint64 (&ulimit),
    GNUNET_PQ_query_param_end
  };
  enum GNUNET_DB_QueryStatus qs;
  struct HistoricReserveRevenueContext hrc = {
    .cb = cb,
    .cb_cls = cb_cls,
    .pg = pg
  };

  PREPARE (pg,
           "iterate_historic_reserve_revenue_summary_select_inc",
           "SELECT"
           " row_id"
           ",start_date"
           ",end_date"
           ",reserve_profits"
           " FROM auditor_historic_reserve_summary"
           " WHERE row_id > $1"
           " ORDER BY row_id ASC"
           " LIMIT $2");
  PREPARE (pg,
           "iterate_historic_reserve_revenue_summary_select_dec",
           "SELECT"
           " row_id"
           ",start_date"
           ",end_date"
           ",reserve_profits"
           " FROM auditor_historic_reserve_summary"
           " WHERE row_id < $1"
           " ORDER BY row_id DESC"
           " LIMIT $2");
  qs = GNUNET_PQ_eval_prepared_multi_select (
    pg->conn,
    limit > 0
    ? "iterate_historic_reserve_revenue_summary_select_inc"
    : "iterate_historic_reserve_revenue_summary_select_dec",
    params,
    &historic_reserve_revenue_cb,
    &hrc);
  if (0 >= qs)
    return qs;
  return hrc.qs;
}
